;***** LED Fader version 2. PWM include file
;      Copyright (c) 2003, Mikael Ejberg Pedersen
;      This program is free for personal use.
;
;	Last edited for version 2.01


.equ	PWMPorts	=2
.equ	PWMLevels	=64	; NOT to be changed (without rewriting a lot of the code)

.equ	PWMPort1	=PORTB
.equ	PWMPort2	=PORTA

.equ	PWMChannels	=16


.dseg

PWMTable:	.byte PWMChannels
OutputTable:	.byte PWMPorts * PWMLevels


.cseg

PWM_OC1A_int:
	push	w1
	in	w1,SREG
	push	w1
	push	w2
	push	ZL
	push	ZH
	push	r0
	push	r1
	
	inc	PWMCount
	ldi	w1,PWMLevels-1
	and	PWMCount,w1		; Only use 0-63. This way 64 will be fully on
	cp	PWMCount,w1		; Last interrupt before new PWM loop?
	brne	OC1A_NotNewLoop
	sbr	Flags,1			; Signal calculation of new PWM values
OC1A_NotNewLoop:
	
	ldi	ZL,low(TC1TopTable*2)
	ldi	ZH,high(TC1TopTable*2)
	mov	w1,PWMCount
	clr	w2
	lsl	w1			; *2 (word count)
	rol	w2
	add	ZL,w1
	adc	ZH,w2
	lpm	w1,Z+
	lpm	w2,Z+
	out	OCR1AH,w2		; Set TC1TOP
	out	OCR1AL,w1
	
	ldi	ZL,low(OutputTable)
	ldi	ZH,high(OutputTable)
	ldi	w1,PWMPorts
	mul	w1,PWMCount
	add	ZL,r0
	adc	ZH,r1
	ld	w1,Z+
	com	w1			; REMOVE when output isn't inverted
	out	PWMPort1,w1
	ld	w1,Z+
	com	w1			; REMOVE when output isn't inverted
	out	PWMPort2,w1
					; Add ld and out here for more ports
	pop	r1
	pop	r0
	pop	ZH
	pop	ZL
	pop	w2
	pop	w1
	out	SREG,w1
	pop	w1
	reti
;66 cycles (+2 for 2 COM)

TC1TopTable:
	.dw	  80,  80,  80,  80,  80,  80,  80,  80
	.dw	  80,  80, 105, 258, 302, 351, 402, 458
	.dw	 517, 580, 646, 715, 789, 866, 946,1031
	.dw	1118,1209,1304,1403,1504,1610,1719,1832
	.dw	1948,2068,2192,2318,2449,2583,2721,2862
	.dw	3008,3155,3308,3463,3623,3785,3952,4121
	.dw	4296,4472,4653,4837,5025,5217,5411,5610
	.dw	5812,6018,6227,6440,6657,6876,7101,7327

.equ	InitialTC1TOP = 7327	; For a long startup to avoid stressing bootsequence

PWM_Init1:
	ser	w1
	out	PWMPort1-1,w1		; -1 = DDRx
	out	PWMPort2-1,w1		; Add more ports here
	clr	w1			; ENABLE when output isn't inverted
	out	PWMPort1,w1
	out	PWMPort2,w1		; Add more ports here

	ldi	ZL,low(OutputTable)
	ldi	ZH,high(OutputTable)
	clr	w1
	ldi	w2,PWMPorts*PWMLevels
PWMI_ClearOutputTable:
	st	Z+,w1
	dec	w2
	brne	PWMI_ClearOutputTable
	
	ldi	ZL,low(PWMTable)
	ldi	ZH,high(PWMTable)
	clr	w1
	ldi	w2,PWMChannels
PWMI_ClearPWMTable:
	st	Z+,w1
	dec	w2
	brne	PWMI_ClearPWMTable
	ret
	

PWM_Init2_Timer1setup:
	ldi	w1,high(InitialTC1TOP)		;Init Timer/Counter 1
	out	OCR1AH,w1
	ldi	w1,low(InitialTC1TOP)
	out	OCR1AL,w1
	clr	w1
	out	TCCR1A,w1
	ldi	w1,(1<<WGM12)|(1<<CS10)	;Mode 4 (CTC), no prescaler
	out	TCCR1B,w1
	ldi	w1,(1<<OCIE1A)		;Enable OCR1A interrupt
	out	TIMSK,w1
	ret


PWMToOutput_WXY:
	ldi	XL,low(OutputTable)
	ldi	XH,high(OutputTable)
	ser	w1
	ser	w2
	clr	w5			; PWM Counter
PWMTO_OuterLoop:
	ldi	YL,low(PWMTable)
	ldi	YH,high(PWMTable)
	ldi	w3,0xfe			; Bitmask
PWMTO_InnerLoop:
	ldd	w4,Y+8
	cp	w4,w5
	brne	PWMTO_NotZero1
	and	w2,w3
PWMTO_NotZero1:
	ld	w4,Y+
	cp	w4,w5
	brne	PWMTO_NotZero2
	and	w1,w3
PWMTO_NotZero2:
	sec
	rol	w3
	brcs	PWMTO_InnerLoop
	st	X+,w1
	st	X+,w2
	inc	w5
	cpi	w5,PWMLevels
	brne	PWMTO_OuterLoop
	ret

