;***** LED Fader version 2. Script command include file
;      Copyright (c) 2003, Mikael Ejberg Pedersen
;      This program is free for personal use.
;
;	Last edited for version 2.01


.equ	MaxConcurrentScripts=10
.equ	MaxScriptStack=3


.equ	ScriptRecPtr	=0	; word
.equ	ScriptRecDelay	=2	; word
.equ	ScriptRecOffset	=4
.equ	ScriptRecStackPtr=5
.equ	ScriptRecStack	=6	; 2*MaxScriptStack
.equ	ScriptRecSize	=6+(2*MaxScriptStack)


.dseg
ScriptRec:	.byte MaxConcurrentScripts * ScriptRecSize


.cseg

Script_Init:
	ldi	ZL,low(ScriptRec)
	ldi	ZH,high(ScriptRec)
	clr	w1
	ldi	w2,MaxConcurrentScripts*ScriptRecSize
SI_ClearScriptRecs:
	st	Z+,w1
	dec	w2
	brne	SI_ClearScriptRecs

	ldi	YL,low(ScriptRec)	; Init first script
	ldi	YH,high(ScriptRec)
	ldi	w1,low(ScriptStart*2)
	ldi	w2,high(ScriptStart*2)
	std	Y+ScriptRecPtr,w1
	std	Y+ScriptRecPtr+1,w2
	clr	w1
	std	Y+ScriptRecDelay,w1	; No delay
	std	Y+ScriptRecDelay+1,w1
	std	Y+ScriptRecOffset,w1	; No offset
	std	Y+ScriptRecStackPtr,w1	; No stack in use
	ret



RunScripts_WXYZ:
	ldi	w1,MaxConcurrentScripts
	mov	ScriptCount,w1
	ldi	YL,low(ScriptRec)
	ldi	YH,high(ScriptRec)
RS_Loop:
	ldd	ZL,Y+ScriptRecPtr	; Load ScriptPointer
	ldd	ZH,Y+ScriptRecPtr+1
	mov	w1,ZL
	or	w1,ZH			; ScriptPointer=0 ?
	breq	RS_NextScript
	ldd	XL,Y+ScriptRecDelay	; Load ScriptDelay
	ldd	XH,Y+ScriptRecDelay+1
	mov	w1,XL
	or	w1,XH			; Delay=0 ?
	brne	RS_Delay
	lpm	w1,Z+			; Load function call address
	lpm	w2,Z+
	movw	XH:XL,ZH:ZL		; Move scriptaddress to X
	movw	ZH:ZL,w2:w1		; Move function address til Z
	push	YL
	push	YH
	icall				; Call scriptfunction. Z points at first parameter
					; Y points at the script record
	pop	YH
	pop	YL
	std	Y+ScriptRecPtr,ZL	; Save scriptpointer
	std	Y+ScriptRecPtr+1,ZH
	rjmp	RS_Loop			; Run next scriptline, delay or whatever
RS_Delay:
	sbiw	XH:XL,1			; Countdown delay
	std	Y+ScriptRecDelay,XL	; Save new delay
	std	Y+ScriptRecDelay+1,XH
RS_NextScript:
	adiw	YH:YL,ScriptRecSize	; Find next scriptrecord
	dec	ScriptCount
	brne	RS_Loop			; More concurrent scripts
	ret


;--------------------------------------------------------------------------
; SCRIPT FUNCTIONS
;
; All scripts MUST return pointer to next scriptline in Z
; (right after last parameter)
; Y points at the current scriptrecord, in case it is needed
; All registers (WXYZ) may be used, but Z MUST return above mentioned value
;--------------------------------------------------------------------------

;Set: Set a channel directly to a value without fading
;Usage: .dw S_Set, Channel(relative), Value
; Channel	0-15
; Value		0-64			; Yes... I DO mean that 64 is included
S_Set:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	ldd	w2,Y+ScriptRecOffset	; Get LEDOffset before Y is destroyed
	lpm	w1,Z+			; Read channel
	adiw	ZH:ZL,1			; Word alignment
	add	w1,w2			; Add offset
	ldi	YL,low(LedRec)
	ldi	YH,high(LedRec)
	ldi	w2,LedRecSize
	mul	w1,w2
	add	YL,r0
	adc	YH,r1
	lpm	w1,Z+			; Read value
	adiw	ZH:ZL,1			; Word alignment
	std	Y+LedRecCurrent,w1
	std	Y+LedRecFinal,w1
	ret


;Increment: Increments a channel by one
;Usage: .dw S_Increment, Channel(relative)
; Channel	0-15
S_Increment:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	ldd	w2,Y+ScriptRecOffset	; Get LEDOffset before Y is destroyed
	lpm	w1,Z+			; Read channel
	adiw	ZH:ZL,1			; Word alignment
	add	w1,w2			; Add offset
	ldi	YL,low(LedRec)
	ldi	YH,high(LedRec)
	ldi	w2,LedRecSize
	mul	w1,w2
	add	YL,r0
	adc	YH,r1
	ldd	w1,Y+LedRecCurrent
	inc	w1
	cpi	w1,PWMLevels+1
	brlo	S_I_NotHigh
	ldi	w1,PWMLevels
S_I_NotHigh:
	std	Y+LedRecCurrent,w1
	std	Y+LedRecFinal,w1
	ret

	
;Decrement: Decrements a channel by one
;Usage: .dw S_Decrement, Channel(relative)
; Channel	0-15
S_Decrement:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	ldd	w2,Y+ScriptRecOffset	; Get LEDOffset before Y is destroyed
	lpm	w1,Z+			; Read channel
	adiw	ZH:ZL,1			; Word alignment
	add	w1,w2			; Add offset
	ldi	YL,low(LedRec)
	ldi	YH,high(LedRec)
	ldi	w2,LedRecSize
	mul	w1,w2
	add	YL,r0
	adc	YH,r1
	ldd	w1,Y+LedRecCurrent
	dec	w1
	brpl	S_I_NotHigh
	clr	w1
S_D_NotLow:
	std	Y+LedRecCurrent,w1
	std	Y+LedRecFinal,w1
	ret
	

;Fade: Set up a channel to fade up or down to a certain value
;Usage: .dw S_Fade, Channel(relative), FinalValue, Duration
; Channel	0-15
; FinalValue	0-64			; Yes... I DO mean that 64 is included
; Duration	1-65535
S_Fade:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	ldd	w2,Y+ScriptRecOffset	; Get LEDOffset before Y is destroyed
	lpm	w1,Z+			; Read channel
	adiw	ZH:ZL,1			; Word alignment
	add	w1,w2			; Add offset
	ldi	YL,low(LedRec)
	ldi	YH,high(LedRec)
	ldi	w2,LedRecSize
	mul	w1,w2
	add	YL,r0
	adc	YH,r1
	lpm	w1,Z+			; Read final value
	adiw	ZH:ZL,1			; Word alignment
	std	Y+LedRecFinal,w1
	lpm	w1,Z+			; Read duration
	lpm	w2,Z+
	std	Y+LedRecDuration,w1
	std	Y+LedRecDuration+1,w2
	call	CalculateLedFadeParams_W12X
	ret


;Delay delays current scriptthread for Param * 10ms
;Usage: .dw S_Delay, Delay	(Delay=100 for a 1 second delay)
; Delay		1-65535
S_Delay:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	lpm	XL,Z+
	lpm	XH,Z+
	std	Y+ScriptRecDelay,XL
	std	Y+ScriptRecDelay+1,XH
	ret


;Goto jumps scriptexecution to another place
;Usage: .dw S_Goto, NewLabel
S_Goto:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	lpm	XL,Z+			; Get new pointer
	lpm	XH,Z+
	movw	ZH:ZL,XH:XL
	lsl	ZL			; *2
	rol	ZH
	ret


;Gosub calls a "subscript"
;Usage: .dw S_Gosub, SubLabel
S_Gosub:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	ldd	w1,Y+ScriptRecStackPtr	; Get stackpointer
	cpi	w1,MaxScriptStack
	brsh	S_G_Error		; Stack already full. Can't gosub
	mov	w2,w1
	inc	w2
	std	Y+ScriptRecStackPtr,w2	; Save new stackpointer
	lsl	w1			; StackPointer*2
	clr	w2
	adiw	XH:XL,2			; ScriptPointer already in X. Add 2 to skip parameters
	add	YL,w1			; Add stackpointerindex
	adc	YH,w2
	std	Y+ScriptRecStack,XL	; Save returnaddress on stack
	std	Y+ScriptRecStack+1,XH
	lpm	XL,Z+			; Get new pointer (to subscript)
	lpm	XH,Z+
	movw	ZH:ZL,XH:XL
	lsl	ZL			; *2
	rol	ZH
	rjmp	S_G_Exit
S_G_Error:
	adiw	ZH:ZL,2			; Skip parameters
S_G_Exit:
	ret


;Return returns to previous called Gosub
;Usage: .dw S_Return	(no parameters)
S_Return:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	ldd	w1,Y+ScriptRecStackPtr	; Get stackpointer
	tst	w1			; Stackpointer=0 ?
	breq	S_R_Error		; Stack empty. Can't return
	dec	w1
	std	Y+ScriptRecStackPtr,w1	; Save new stackpointer
	lsl	w1			; StackPointer*2
	clr	w2
	add	YL,w1			; Add stackpointerindex
	adc	YH,w2
	ldd	ZL,Y+ScriptRecStack	; Load returnaddress from stack
	ldd	ZH,Y+ScriptRecStack+1
	rjmp	S_R_Exit
S_R_Error:
	clr	ZL			; Clear scriptpointer to stop thread
	clr	ZH
S_R_Exit:
	ret


;Start initiates a new script thread
;Usage: .dw S_Start, NewScript, Offset
S_Start:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	lpm	XL,Z+			; Load new script startaddress
	lpm	XH,Z+
	lsl	XL			; *2
	rol	XH
	lpm	w1,Z+			; Load offset
	adiw	ZH:ZL,1			; Word alignment
	ldi	w2,MaxConcurrentScripts
	ldi	YL,low(ScriptRec)
	ldi	YH,high(ScriptRec)
S_S_Loop:
	ldd	w3,Y+ScriptRecPtr	; Load ScriptPointer
	ldd	w4,Y+ScriptRecPtr+1
	or	w3,w4			; ScriptPointer=0 ?
	breq	S_S_FreeRec
	adiw	YH:YL,ScriptRecSize	; Find next scriptrecord
	dec	w2
	brne	S_S_Loop		; More scriptrecords
	rjmp	S_S_Exit
S_S_FreeRec:
	std	Y+ScriptRecPtr,XL	; Save new scriptpointer
	std	Y+ScriptRecPtr+1,XH
	clr	w2
	std	Y+ScriptRecDelay,w2	; Clear ScriptDelay
	std	Y+ScriptRecDelay+1,w2
	std	Y+ScriptRecOffset,w1	; Set offset
	std	Y+ScriptRecStackPtr,w2	; No stack in use
S_S_Exit:
	ret


;Stop stops the current script thread and releases it for later use
;Usage: .dw S_Stop	(no parameters)
S_Stop:
	clr	ZL			; Clear scriptpointer to stop thread
	clr	ZH
	ret


;BranchIfBitSet branches to a new location if a bit in a memorylocation is set
;Usage: .dw S_BranchIfBitSet, MemAddr, Bit, Label
S_BranchIfBitSet:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	lpm	XL,Z+			; Load address of byte to bittest
	lpm	XH,Z+
	ld	w1,X			; Load byte to bittest
	lpm	w2,Z+			; Load bit number to test
	adiw	ZH:ZL,1			; Word alignment
	lpm	XL,Z+			; Load new script address (in case it's needed)
	lpm	XH,Z+
	ldi	w3,1
S_BIBS_BitLoop:
	tst	w2
	breq	S_BIBS_BitLoopExit
	dec	w2
	lsl	w3
	rjmp	S_BIBS_BitLoop
S_BIBS_BitLoopExit:
	and	w1,w3
	breq	S_BIBS_Exit		; Bit was cleared: Continue without branching
	movw	ZH:ZL,XH:XL		; Bit was set: Set new address
	lsl	ZL			; *2
	rol	ZH
S_BIBS_Exit:
	ret


;BranchIfBitClear branches to a new location if a bit in a memorylocation is cleared
;Usage: .dw S_BranchIfBitClear, MemAddr, Bit, Label
S_BranchIfBitClear:
	movw	ZH:ZL,XH:XL		; Restore Z (pointer to first parameter)
	lpm	XL,Z+			; Load address of byte to bittest
	lpm	XH,Z+
	ld	w1,X			; Load byte to bittest
	lpm	w2,Z+			; Load bit number to test
	adiw	ZH:ZL,1			; Word alignment
	lpm	XL,Z+			; Load new script address (in case it's needed)
	lpm	XH,Z+
	ldi	w3,1
S_BIBC_BitLoop:
	tst	w2
	breq	S_BIBC_BitLoopExit
	dec	w2
	lsl	w3
	rjmp	S_BIBC_BitLoop
S_BIBC_BitLoopExit:
	and	w1,w3
	brne	S_BIBC_Exit		; Bit was set: Continue without branching
	movw	ZH:ZL,XH:XL		; Bit was cleared: Set new address
	lsl	ZL			; *2
	rol	ZH
S_BIBC_Exit:
	ret

